import unittest
from unittest import mock
import commands
from click.testing import CliRunner

class TestSaveCommand(unittest.TestCase):

    @mock.patch('commands.repo.git.commit')
    @mock.patch('commands.get_conflict_files')
    @mock.patch('commands.repo.remotes.origin.exists')
    @mock.patch('commands.repo.git.pull')
    @mock.patch('commands.conflicts_from_merge')
    @mock.patch('commands.repo.git.push')
    def test_save_commit_exception(
        self,
        mock_get_conflict_files, 
        mock_commit,
        mock_exists,
        mock_pull,
        mock_conflicts_from_merge,
        mock_push
        ):
        mock_get_conflict_files.return_value = False
        mock_commit.side_effect = Exception()
        mock_exists.return_value = True
        mock_pull.return_value = True
        runner = CliRunner()
        result = runner.invoke(commands.save, 
        ['-m', 'message'])
        self.assertTrue(mock_push.called)

    @mock.patch('commands.repo.git.commit')
    @mock.patch('commands.get_conflict_files')
    @mock.patch('commands.repo.remotes.origin.exists')
    @mock.patch('commands.repo.git.pull')
    @mock.patch('commands.conflicts_from_merge')
    @mock.patch('commands.repo.git.push')
    @mock.patch('commands.click.echo')
    def test_save_conflicted_files(
        self,
        mock_get_conflict_files, 
        mock_commit,
        mock_exists,
        mock_pull,
        mock_conflicts_from_merge,
        mock_echo
        ):
        mock_get_conflict_files.return_value = True
        mock_commit.side_effect = Exception()
        mock_exists.return_value = True
        mock_pull.return_value = True
        runner = CliRunner()
        result = runner.invoke(commands.save, 
        ['-m', 'message'])
        mock_echo.assert_called_with('Please fix the following conflicts then use "gt save" again')


import os
import git
import click
import requests
from datetime import datetime

from goBack.commands import goBack
from init.commands import init
from save.commands import save
from ignore.commands import ignore
from compare.commands import compare
from lesson.commands import lesson


class CustomGroup(click.Group):
    def invoke(self, ctx):
        ctx.obj['args'] = tuple(ctx.args)
        super(CustomGroup, self).invoke(ctx)

@click.group(cls=CustomGroup)
@click.pass_context
def cli(ctx):
    """ 
    Git the easy way.

    If you want to access gitutor tutorials run 

        $ gt lesson

    Any issues, questions or bugs you can reach us at support@gitutor.io
    """
    #Check ctx was initialized
    ctx.ensure_object(dict)

    
    if ('--help' not in ctx.obj['args']) and (ctx.invoked_subcommand != 'init') and ctx.invoked_subcommand != 'lesson':
        try:
            repo = git.Repo(".", search_parent_directories=True)
            ctx.obj['REPO'] = repo
            #print( f"Location {repo.working_tree_dir}" )
            #print(f"Remote from init: {repo.remote('origin').url} ")
        except Exception as e:
            click.echo('Ups! You\'re not inside a git repo')
            #print("not git repo")
            exit()

cli.add_command(init)
cli.add_command(goBack)
cli.add_command(save)
cli.add_command(ignore)
cli.add_command(compare)
cli.add_command(lesson)

def main():
    cli(obj={})

if __name__ == '__main__':
    cli(obj={})



import git
import click
from git import GitCommandError
from .color_diff import color_diff
from util import commits_full_list, prompt_for_commit_selection

@click.command(short_help = "Compare current status with selected commit")
@click.pass_context
@click.option('-h', '--hash', 'commit', help='Hash of commit to compare')
def compare(ctx,commit):
    """
    Compare current status with selected commit.

    To display a list with all the commits run:

        $ gt compare 

    If you know the hash of the commit you want to compare your repo with, then run
    
        $ gt compare --hash  <commitHash>

    """
    #Recover repo from context
    repo = ctx.obj['REPO']

    if commit:
        try:
            click.echo(repo.git.diff(commit))
        except GitCommandError:
            click.echo('Hash not found')
        return


    full_list_of_commits = commits_full_list(repo)
    click.echo('\n'.join(color_diff("-Red lines are from prior version. \n+Green lines are from current version".splitlines())))
    click.echo("Learn how to read diff output here: https://gitutor.io/guide/gt-compare.html#understanding-the-output \n")
    answer = prompt_for_commit_selection(full_list_of_commits, 'Select commit to compare')
    if answer:
        answer_hash = answer['commit'][:7]
        diff_string = repo.git.diff(answer_hash)
        click.echo('\n'.join(color_diff(diff_string.splitlines())))



try:
    from colorama import Fore, Back, Style, init
    init()
except ImportError:  # fallback so that the imported classes always exist
    class ColorFallback():
        __getattr__ = lambda self, name: ''
    Fore = Back = Style = ColorFallback()

def color_diff(diff):
    for line in diff:
        if line.startswith('+'):
            yield Fore.GREEN + line + Fore.RESET
        elif line.startswith('-'):
            yield Fore.RED + line + Fore.RESET
        elif line.startswith('@'):
            yield Fore.BLUE + line + Fore.RESET
        else:
            yield line


import unittest
from unittest import mock
import commands
from click.testing import CliRunner
from git import Repo

class TestGoBackCommand(unittest.TestCase):

    # @mock.patch('commands.git')
    # def test_is_initialized_true(self, mock_git):
    #     is_initialized = commands.is_initialized_repo()
    #     self.assertTrue(is_initialized)

    @mock.patch('commands.revert_hash_sequence')
    @mock.patch('commands.hash_sequence_to_target')
    def test_revert_to_target_empty_hash_sequence(
        self,
        mock_hash_seq_to_target,
        mock_revert_hash_seq
    ):
        repo = mock.Mock(spec=Repo)
        mock_hash_seq_to_target.side_effect = [
            [1],
            []
        ]
        number_of_reverts = commands.revert_to_target(repo, 1, [])
        self.assertFalse(mock_revert_hash_seq.called)
        self.assertFalse(repo.git.called)
        self.assertTrue(mock_hash_seq_to_target.called)
        self.assertEqual(number_of_reverts, 0)

    @mock.patch('commands.revert_hash_sequence')
    @mock.patch('commands.hash_sequence_to_target')
    def test_revert_to_target_succesfull(
        self,
        mock_hash_seq_to_target,
        mock_revert_hash_seq
    ):
        repo = mock.Mock(spec=Repo)
        mock_hash_seq_to_target.side_effect = [
            [1, 2, 3],
            [1, 2, 3]
        ]
        number_of_reverts = commands.revert_to_target(repo, 1, [])
        self.assertTrue(mock_revert_hash_seq.called)
        self.assertFalse(repo.git.called)
        self.assertTrue(mock_hash_seq_to_target.called)
        self.assertEqual(number_of_reverts, 3)

    @mock.patch('commands.revert_hash_sequence')
    @mock.patch('commands.hash_sequence_to_target')
    def test_revert_to_target_fails(
        self,
        mock_hash_seq_to_target,
        mock_revert_hash_seq
    ):
        repo = mock.Mock(spec=Repo)
        mock_hash_seq_to_target.side_effect = [
            [1, 2, 3],
            [1, 2, 3]
        ]
        mock_revert_hash_seq.side_effect = Exception
        number_of_reverts = commands.revert_to_target(repo, 1, [])
        self.assertTrue(mock_revert_hash_seq.called)
        self.assertTrue(mock_hash_seq_to_target.called)
        repo.git.reset.assert_called_with('--hard', 1)
        self.assertEqual(number_of_reverts, 0)

    


import click
from PyInquirer import prompt
from .lessons import lessons, welcome_message

@click.command()
@click.pass_context
def lesson(ctx):
    """ 
    Gitutor lessons right on your terminal !
    """
    click.echo(welcome_message)
    selection = prompt_for_lesson()
    if selection:
        lesson_name = extract_lesson_name(selection['lesson'])
        click.echo(lessons[lesson_name]['content'])


def create_choices_from_lessons():
    return [f'{lesson} - {lessons[lesson]["description"]}' for lesson in lessons]

def extract_lesson_name(selection):
    if selection:
        return selection.split('-')[0].strip()

def prompt_for_lesson():
    message = 'Select a lesson!'
    choices = create_choices_from_lessons()
    question = [
        {
            'type': 'list',
            'message': message,
            'name': 'lesson',
            'choices': choices
        }
    ]
    return prompt(question)


import git
import click
from git import GitCommandError

@click.command(short_help = "Creates a checkpoint of your project")
@click.pass_context
@click.option('-m', '--message', 'message', help='Commit message')
@click.option('-l', '--local', is_flag=True, help='Save locally only')
@click.option('--defer-conflicts', is_flag=True, help='Defer conflict resolution')
def save(ctx, message, local, defer_conflicts):
    """
    Creates a checkpoint of your project, i.e., saves all of your current changes and new files.

    You can pass the commit message without the prompt with:

        $ gt save --message "This is a commit"

    If you don't want to push your changes to the remote repository run:

        $ gt save --local

    If you have files with conflicts but don't want to resolve them right away run: 

        $ gt save --defer-conflicts

    This will abort the attempt to sync with the remote repo. Your changes will be saved on your local machine only.
    You will have to solve the merge conflict later on.
    """
    # Recover repo from context
    repo = ctx.obj['REPO']

    if defer_conflicts:
        try:
            repo.git.merge('--abort')
        except Exception as e:
            if e.args[1] == 128:
                click.echo('No merge to abort!')
        else:
            click.echo('Merge aborted!')
        return

    conflicted_files = get_conflict_files(repo)

    if not conflicted_files:

        #git add .
        if not message:
            message = click.prompt('Commit message')
    
        repo.git.add(A=True)

        #git commit -m ""
        try:
            repo.git.commit("-m", message)
        except GitCommandError:
            click.echo("No changes in local repository since last commit")

        if not local:
            try: 
                repo.remotes.origin.exists()
            except:
                click.echo('Remote repository does not exist!')
            else: 
                try:
                    repo.git.pull()
                except GitCommandError as e:
                    if "no tracking" in e.stderr:
                        click.echo("There is no corresponding branch in the remote repository")
                        click.echo("Changes were saved only in local repository")
                    else:
                        conflicted_files_merge = conflicts_from_merge(repo)
                        if conflicted_files_merge:
                            click.echo('Merge conflicts in:')
                            for conflict_file in conflicted_files_merge: 
                                click.echo("- " + conflict_file)
                            click.echo('Please fix conflicts then use "gt save" again')    
                            print_abort_merge_instructions()
                else:
                    click.echo('Pushing to remote repository...')
                    repo.git.push()
                    click.echo('Done!')
    else:
        click.echo('Please fix the following conflicts then use "gt save" again')
        for conflicted_file in conflicted_files:
            click.echo(conflicted_file)
        print_abort_merge_instructions()

def print_abort_merge_instructions():
        click.echo('\nLearn how to resolve conflicts here: https://gitutor.io/guide/multiple-users.html#merge-conflicts')
        click.echo('\nDon\'t want conflicts right now ?')
        click.echo('Defer conflict resolution: gt save --defer-conflicts ')
        click.echo('Dont worry, your work will still be saved on your local repo:)')
        click.echo('\nSave only on local repo until you want to resolve conflict: gt save -l ')

def conflicts_from_merge(repo):
    unmerged_blobs = repo.index.unmerged_blobs()
    error_array = []
    # We're really interested in the stage each blob is associated with.
    # So we'll iterate through all of the paths and the entries in each value
    # list, but we won't do anything with most of the values.
    for path in unmerged_blobs:
        list_of_blobs = unmerged_blobs[path]
        for (stage, blob) in list_of_blobs:
            # Now we can check each stage to see whether there were any conflicts
            if stage != 0:
                if path not in error_array:
                    error_array.append(path)
    return error_array

def get_conflict_files(repo):
    matched_lines=[]
    try:
        repo.git.diff("--check")
    except GitCommandError as e:
        matched_lines = [line for line in e.stdout.split("'")[1].split('\n') if (": leftover conflict marker" in line) ]
    return matched_lines




from PyInquirer import prompt
import click

def prompt_for_commit_selection(list_of_commits, message):
    commit_page_index = 0
    choices = get_commit_page(commit_page_index, list_of_commits)
    question = [
        {
            'type': 'list',
            'message': message,
            'name': 'commit',
            'choices': choices
        }
    ]
    answer = prompt(question)
    if answer:
        while answer['commit'] == '...Show previous commits' or answer['commit'] == 'Show more commits...':
            if answer['commit'] == '...Show previous commits':
                commit_page_index-=1
                commits_to_show=get_commit_page(commit_page_index, list_of_commits)
            else:
                commit_page_index+=1
                commits_to_show=get_commit_page(commit_page_index, list_of_commits)
            question[0].update({'choices': commits_to_show})
            answer = prompt(question)
    return answer

def commits_full_list(repo, first_parent = True):
    
    if first_parent:
        log_params = ['--first-parent']
    else:
        log_params = []
    
    log_params.append("--pretty=format:'%h - %s: %an, %ar'")
    full_list_of_commits = repo.git.log(log_params).replace("'","").splitlines()
    return full_list_of_commits

def get_commit_page(page_index, full_list_of_commits):
    if page_index == 0:
        commits_to_show = full_list_of_commits[0:20]
    else:
        commits_to_show = full_list_of_commits[page_index*20:page_index*20+20]
        commits_to_show.insert(0,'...Show previous commits')
    if len(full_list_of_commits) > page_index*20+20:    
        commits_to_show.append('Show more commits...')
    return commits_to_show


import unittest
from unittest import mock
import commands
from click.testing import CliRunner

class TestInitCommand(unittest.TestCase):

    @mock.patch('commands.git')
    def test_is_initialized_true(self, mock_git):
        is_initialized = commands.is_initialized_repo()
        self.assertTrue(is_initialized)

    @mock.patch('commands.git')
    def test_is_initialized_false(self, mock_git):
        mock_git.Repo.side_effect = Exception()
        is_initialized = commands.is_initialized_repo()
        self.assertFalse(is_initialized)

    @mock.patch('commands.Github')
    @mock.patch('commands.init_local_repo')
    @mock.patch('commands.is_initialized_repo')
    def test_init_github_exception(
        self,
        mock_is_initialized_repo, 
        mock_init_local_repo, 
        mock_Github
        ):
        mock_is_initialized_repo.return_value = False
        mock_Github.side_effect = Exception()
        runner = CliRunner()
        result = runner.invoke(commands.init, 
        ['-u', 'user', '-p', 'passwrod', '-r', 'repo_name'])
        self.assertFalse(mock_init_local_repo.called)

    @mock.patch('commands.Github')
    @mock.patch('commands.init_local_repo')
    @mock.patch('commands.is_initialized_repo')
    def test_init_already_initialized(
        self,
        mock_is_initialized_repo, 
        mock_init_local_repo, 
        mock_Github
        ):
        mock_is_initialized_repo.return_value = True
        runner = CliRunner()
        result = runner.invoke(commands.init, 
        ['-u', 'user', '-p', 'passwrod', '-r', 'repo_name'])
        self.assertFalse(mock_Github.called)
        self.assertFalse(mock_init_local_repo.called)

    @mock.patch('commands.Github')
    @mock.patch('commands.init_local_repo')
    @mock.patch('commands.is_initialized_repo')
    def test_init_local_only(
        self,
        mock_is_initialized_repo, 
        mock_init_local_repo, 
        mock_Github
        ):
        mock_is_initialized_repo.return_value = False
        runner = CliRunner()
        result = runner.invoke(commands.init, 
        ['-l'])
        self.assertFalse(mock_Github.called)
        self.assertTrue(mock_init_local_repo.called)



import os
import re
import click
import difflib
from PyInquirer import prompt
from git import GitCommandError

def get_files(path:str, git_path:str):
    '''
    Get all files in path, including hidden files
    '''
    files = os.listdir(path)

    # remove git files from list
    for f in ['.git', '.gitignore', '.github']:
        try:
            files.remove(f)
        except:
            pass
    
    # get dir relative path to git's repo
    relative_path = ''.join([x[-1] for x in difflib.ndiff(path, git_path) if x[0] == '-'])
    relative_path = re.sub('^/', '', relative_path)

    # create dictionary with files (key is the plain file name)
    dir_files = {}
    for key in files:
        full_path = '{}/{}'.format(path,key)
        folder = os.path.isdir(full_path)

        if folder: key = key + '/'

        dir_files[key] = {
            'file_name': '{}/{}'.format(relative_path, key) if len(relative_path) > 1 else key,
            'is_folder': folder
        }

    return dir_files

def get_gitignore_files(git_path:str):  
    '''
    Get the files that already are in the .gitignore file
    '''
    gitignore_path = git_path + '/.gitignore'
    files = []

    if os.path.exists(gitignore_path):
        with open(gitignore_path, 'r') as _:
            files = re.sub(r'(\n+$)|(^\n)', '', _.read())

        files = re.sub(r'(\n\s*){2,}','\n', files)
        files = files.split("\n")

    return files

def create_choices(dir_files:dict, gitignore_files:list):
    '''
    Create choices list for the interactive menu.
    Files in .gitignore are checked
    '''
    choices = []

    for file in dir_files:
        choice = {
            'name': file
        }

        if dir_files[file]['file_name'] in gitignore_files:
            choice['checked'] = True
            dir_files[file]['is_ignored'] = True
        else:
            dir_files[file]['is_ignored'] = False

        choices.append(choice)

    return choices

def update_gitignore(answers:list, dir_files:dict, gitignore_files:list, git_path: str):
    '''
    Update gitignore file
    '''
    new_files = []

    # add new files to gitignore
    for file in answers:
        if dir_files[file]['file_name'] not in gitignore_files:
            gitignore_files.append(dir_files[file]['file_name'])  
            new_files.append(dir_files[file]['file_name'])

    # remove files 
    for key, file_info in filter(lambda d: d[1]['is_ignored'], dir_files.items()):
        if key not in answers:                                                              #file_info['file_name'] not in answers:
            gitignore_files.remove(file_info['file_name'])

    gitignore = '\n'.join(gitignore_files)

    # write gitignore file
    gitignore_path = '{}/.gitignore'.format(git_path)
    with open(gitignore_path, 'w') as _:
        _.write(gitignore)

    return new_files

def remove_tracked_files(new_files:list, repo):
    '''
    Remove tracked files in gitignore from repo
    '''
    tracked_files = repo.git.execute(["git", "ls-files", "-i", "--exclude-standard"])
    tracked_files = tracked_files.split("\n")

    new_tracked_files = []
    for file in new_files:
        if '/' in file:
            regex = re.compile('^'+file+".*")
            new_tracked_files.extend(filter(regex.match, tracked_files))
        elif file in tracked_files:
            new_tracked_files.append(file)
    
    if len(new_tracked_files) > 0:
        repo.git.rm(new_tracked_files, "-r", "--cached")
        try:
            repo.git.commit("-m", "Remove ignored files from repo")
        except GitCommandError as e:
            pass

@click.command()
@click.pass_context
def ignore(ctx):
    """
    Add files to .gitignore with an interactive menu
    """
    # recover repository's object from context
    repo = ctx.obj['REPO']

    # get current ignored files
    git_path = repo.working_tree_dir
    ignored_files = get_gitignore_files(git_path)

    # get files in current directory
    full_path = os.getcwd()
    dir_files = get_files(full_path, git_path)

    questions = [
        {
            'type': 'checkbox',
            'message': 'Select files to ignore',
            'name': 'files',
            'choices': create_choices(dir_files, ignored_files)
        }
    ]

    answers = prompt(questions)

    if 'files' in answers:
        new_ignored_files = update_gitignore(answers['files'], dir_files, ignored_files, git_path)
        remove_tracked_files(new_ignored_files, repo)


import os
import click
from util import prompt_for_commit_selection, commits_full_list
from git import GitCommandError
from collections import deque 
  

@click.command(short_help='Returns repo version to a specific commit')
@click.pass_context
@click.option('-h', '--hash', 'commit_hash', help="hash of the commit to go back")
def goBack(ctx, commit_hash):
    """
    Returns version to a specific commit via hash or selected commit

    To display a list with all the commits run:

        $ gt goback 

    If you know the hash of the commit you want to go back to, then run

        $ gt goback --hash  <commitHash>
    """
    repo = ctx.obj['REPO']

    full_list_of_commits = commits_full_list(repo)
    if commit_hash:
        commit_hash = commit_hash[0:7]
        commit_info = find_commit_info(full_list_of_commits, commit_hash)
    else:
        selected_log_entry = prompt_for_commit_selection(full_list_of_commits, 'Select the commit you want to return')
        if 'commit' not in selected_log_entry: #handle PyInquirer bug that returns None on user click on list
            return
        selected_log_entry = selected_log_entry['commit']
        commit_info = extract_info_from_log_entry(selected_log_entry)

    if not commit_info:
        click.echo('There is no commit with the hash provided')
        return

    click.echo(f'Going back to "{commit_info["comment"]}"...')
    click.echo('Saving changes before revert...')
    commit_before_revert(repo)
    number_of_reverts = revert_to_target(repo, commit_info['hash'], full_list_of_commits)
    if number_of_reverts > 0:
        combine_reverts_into_single_commit(repo, number_of_reverts, commit_info['comment'])
        click.echo('Done!')
    else:
        click.echo('Nothing to revert!')


def combine_reverts_into_single_commit(repo, number_of_reverts, commit_comment):
    index_of_first_revert = number_of_reverts 
    repo.git.reset('--soft', f'HEAD~{index_of_first_revert}')
    repo.git.commit('-m', f'Going back to "{commit_comment}"')

def hash_sequence_to_target(target_hash, list_of_commits):
    list_of_commit_dicts = [ extract_info_from_log_entry(log_entry) for log_entry in list_of_commits ]
    list_of_hashes = [ commit_info['hash'] for commit_info in list_of_commit_dicts ]
    return list_of_hashes[:list_of_hashes.index(target_hash)]

def revert_to_target(repo, target_hash, list_of_commits):
    hash_sequence = hash_sequence_to_target(target_hash, list_of_commits)
    head_sha = hash_sequence[0] if len(hash_sequence) > 0 else None
    #We need to recalculate list_of_commits in case there was a new commit created by commit_before_revert
    list_of_commits = commits_full_list(repo)
    hash_sequence = hash_sequence_to_target(target_hash, list_of_commits)
    if len(hash_sequence) > 0:
        try:
            revert_hash_sequence(repo, hash_sequence)
        except Exception as e:
            print(e)
            repo.git.reset('--hard', head_sha)
            return 0
    
    return len(hash_sequence)

def revert_hash_sequence(repo, hash_sequence):

    for commit_hash in hash_sequence:
        try:
            if is_merge_commit(repo, commit_hash):
                repo.git.revert('-m', '1', commit_hash, '--no-edit')
            else:
                repo.git.revert(commit_hash, '--no-edit')
        except Exception as e:
            raise e

def is_merge_commit(repo, commit_hash):
    return len(repo.commit(commit_hash).parents) > 1

def find_commit_info(list_of_commits, commit_hash):
    list_of_commits = [ extract_info_from_log_entry(log_entry) for log_entry in list_of_commits]
    for commit_info in list_of_commits:
        if commit_info['hash'] == commit_hash:
            return commit_info
        
    return None

def extract_info_from_log_entry(log_entry):
    log_entry_parts = log_entry.split('-')
    return {
        'hash': log_entry_parts[0].strip(),
        'comment': log_entry_parts[1].split(':')[0]
    }

def commit_before_revert(repo):
    repo.git.add(A=True)
    try:
        repo.git.commit("-m", 'Saving before revert')
        return 1
    except GitCommandError:
        return 0


welcome_message ='''\
Here is a list of short lessons to get you going. You can read
the extended version of this tutorial on https://gitutor.io/guide/
'''
lessons = {
    'Introduction': {
        'description': 'Learn the basic concepts',
        'content': '''\
The basic concepts of git:

Git is a free and open source distributed version control system

A version control system is a tool that saves checkpoints (called commits) of
a file or a set of files over time. A collection of files tracked by git is
called a repository (or repo).

Git is distributed because it's possible to have copies of a repository on
different machines in which checkpoints are saved independently.
Github is a service for storing repositories that act as a source of truth.

On the next lesson you'll learn the basic workflows for git
Gitutor provides beginner friendly commands that wrap git.
Start by using the simple commands.
On the extended tutorial you'll be shown what happens behind the scenes and
what are the actual git commands being used.
'''
    },
    'Single user workflow': {
        'description': 'One person works on the repo',
        'content': '''\
The first and simplest workflow is a single person working on the repository.

1. Change directory into the root folder of your project

    $cd myProject

2. Initialize your project's folder as a git repo

    $gt init

3. Work on your project

4. Create checkpoint to save your changes

    $gt save

    Repeat from step 3.

Extended lesson: https://gitutor.io/guide/one-branch.html
'''

    },
    'Multiple users workflow': {
        'description': 'Multiple people on the repo',
        'content': '''\
The simplest workflow for colaboration is the following:

1.  Initialize a repository as shown in Single User Workflow
    or download a copy of a repo with the url from the repo's github page
    with the following git command:

    $git clone url-of-github-repo

    For example:

    $git clone https://github.com/artemisa-mx/demoRepository.git

2. Grant or ask for write permisions to the github repo

3. Work on your project

4. Create checkpoint to save changes.

    $gt save

5.  If no conflict occurs go to step 3.

6.  If a merge conflict occurs, gitutor will output the names of the
    conflicted files. Solve the conflicts and then create a checkpoint to save
    the conflict resolution:

    $gt save

    Repeat from step 3

    Else, if you don't want to resolve the conflict right away, defer conflict resolution:

    $gt save --defer-conflicts

    Your checkpoint will only be saved locally.

When creating a checkpoint on step 4, you can alteratively use:

    $gt save -l

This will save the checkpoint only on your local machine and wont download
changes from github, avoiding any conflict. When you are ready to sync your
local and remote repo use the command without the -l flag.


Learn how to grant write permisions and how to resolve conflicts on the
extended version of the lesson : https://gitutor.io/guide/one-branch.html
'''

    },
    'More than a save button': {
        'description': 'Learn the gitutor funcionality',
        'content': '''\
Besides saving your project's development history, gitutor offers more funcionality.

- You can compare your current version with any previous commit with:

    $ gt compare

An interactive menu will appear with a list of the repository’s commits. Select one
of them to display the differences between your current version and that particular commit.

Documentation for this command can be found here: https://gitutor.io/guide/gt-compare.html

- You can go back to a previous version of your project with:

    $ gt goback

An interactive menu will again show a list of the repo's commits. Select one to
return to that previous version. Doing this will create a new commit so you can undo
the going back action and return to where you were before.

Documentation for this command can be found here: https://gitutor.io/guide/gt-goback.html

- You can tell gitutor to ignore certain files and don't keep track of them. 
To achieve this use:

    $ gt ignore

An interactive menu will display all the files and folders inside your current
location, check the ones you don't want tracked by git. If you select a folder,
git won’t track any file or folder inside of it. Uncheck files or folders so git
starts tracking them.

Any ignored files will not be uploaded to the remote repository. Gitutor will automaticaly
remove any file alredy in the remote repository if it was ignored in your local repository. 

Documentation for this command can be found here: https://gitutor.io/guide/gt-ignore.html
'''

    },
    'gt init': {
        'description': 'Learn how the gt init command works',
        'content': '''\
# gt init

When the 'gt init' command is called, gitutor runs the following commands under the hood:

    $ git init

This initializes a hidden folder with the name ".git". Git stores in this folder all the
information concerning commits and the different versions of your project.

Afterwards a README.md file is created and the first commit is done with:

    $ git add .
    $ git commit -m 'First commit'

These git commands will be explained in the 'gt save' lesson.

Next, gitutor creates a new empty repository on your github account. This can be done from 
ithub's website. If the repository is created with the default README.md option, it won’t
be possible to push your local repo.

Once the repo has it's first commit and the github repo is created, a remote is added with: 

    $ git add remote origin https://github.com/user/repo.git

A remote is an url that points to a repository on another machine (in this case on github's
servers).

"Origin" refers to the name assigned to that particular remote repository, it could be a
name of your choosing.

Once the remote is added, the commit is pushed with the -u flag to set this remote as the
upstream:

    $ git push -u

This will set your remote repository as the default place where commits should be pushed
and where you’ll pull the changes from.

## Options

This commands offers an optional flags:

-l
    $ gt init -l
This flag tells gitutor to start only a local repository, so no remote repository will
be created.
'''

    },
    'gt save': {
        'description': 'Learn how the gt save command works',
        'content': '''\
# gt save

This command will execute the following actions under the hood:

First it runs:

    $  git add .

This tells git to include the changes within all the files in your repository to your
next commit. These changes include any modifications to your previously saved files and
any new (untracked) files. 

Next, in order to actually save the changes to the project’s history gitutor runs:

    $ git commit -m “your message” 

You can think of a commit as a checkpoint or a snapshot of your repository’s current
state. It’s important to notice that this command only saves the changes to your local
repository.

Then we need to check if there are any changes in the remote repository with: 

    $ git pull

This updates your local repository with the remote changes. Git will try to merge the
files but if there is a conflict a message will be prompted indicating which files have
the conflict. You’ll need to solve all the conflicts manually and then run gt save again.

Finally, if there are no conflicts, you’ll be able to upload your local commits, or
checkpoints, to your remote repository. To do this, gitutor runs:

    $ git push

As you can see, if you wanted to use git to save your changes in the local and remote
repository, you would have to run these 4 commands every time. Instead “gt save” does
all of this for you with only 1 command.

## Options

This commands offers two optional flags:

-m
    gt save -m "your commit message"
This flag tells gitutor to use the message introduced in the command directly instead
of prompting the user for one.

-l
    gt save -l
This flag tells gitutor to only save the changes in your local repository. This way
nothing is modified in your remote repository.

'''
    },
    'gt compare': {
        'description': 'Learn how the gt compare command works',
        'content': '''\
# gt compare

The "gt compare" command will let you compare the current state of your files with a
commit you previously made with the "gt save" command.

After executing "gt compare", you are prompted with a list of all the previous commits
done to your project. You can navigate the list with the arrow keys and select the
commit you want to compare with the enter key.

When a commit is selected, gitutor runs the following:

    $ git diff <hashOfSelectedCommit>

Gitutor will then show you a string containing all the differences between the current
state of your project and the selected commit. The green text is the one that was added
after the selected commit and the red text is the one deleted after the selected commit. 

If you wanted to do the same using only git, you would have to know the hash of the
commit you want to compare or specify the number of commits behind to compare. Instead
"gt compare" offers a way to easily select the commit you want to compare your project with.

## Options

This commands offers an optional flags:

-h
    $ gt compare -h <hashOfSelectedCommit>
This flag tells gitutor to compare the current state of your project to the commit with
the inputted hash. This command flag is usefull if you already know the hash of the commit
you want to compare instead of going through the whole list of commits.
'''
    },
    'gt goback': {
        'description': 'Learn how the gt goback command works',
        'content': '''\
# gt goback

The "gt goback" command will let you return the state of your files to a commit you
previously made with the "gt save" command.

After executing "gt goback", you are prompted with a list of all the previous commits
done to your project. You can navigate the list with the arrow keys and select the commit
you want your project to go back to.

When a commit is selected, gitutor runs the following:

    $ git diff hashOfSelectedCommit

Gitutor will then show you a string containing all the differences between the current
state of your project and the selected commit. The green text is the one that was added
after the selected commit and the red text is the one deleted after the selected commit. 

If you wanted to do the same using only git, you would have to know the hash of the commit
you want to compare or specify the number of commits behind to compare. Instead "gt compare"
offers a way to easily select the commit you want to compare your project with.

## Options

This commands offers an optional flags:

-c
    $ gt compare hashOfSelectedCommit
This flag tells gitutor to compare the current state of your project to the commit with
the inputted commit. This command flag is usefull if you already know the hash of the
commit you want to compare instead of going through the whole list of commits.
'''
    },
    'gt ignore': {
        'description': 'Learn how the gt ignore command works',
        'content': '''\
# gt ignore

Sometimes there are files inside our local repository that we don't want to upload to
the remote repository, such as keys, passwords or log files. The *.gitignore* file tells
Git which files or folders to ignore. This means that Git won't track any changes made to
those files and that they won't be uploaded to your remote repository.

It's recommended to have a .gitignore file so that when you run

    $ git add .

Git won't upload any unwanted files by mistake. .gitignore  is a plain-text file in which
each line specifies a pattern to ignore. For example: 

* 'hello.*' will match any file or folder whose name begins with hello ('hello.txt',
'hello.log', 'hello.py')

* 'doc/'will match a folder named 'doc' but not a file named 'doc'. When you ignore
a folder, Git will ignore any files or folders inside of it.

* 'main.py' will match any file named exactly 'main.py'.

* '*.csv' will match every csv file in your repository

It's important to note that the .gitignore file specifies intentionally **untracked** files
that Git should ignore. Files already tracked by Git are not affected so if, for example,
you committed a file named 'keys.py' by mistake and later you add this pattern to the
.gitignore file, nothing will happen. That is, the 'keys.py' file will remain in the
remote repo and Git will still track any modifications. 

Gitutor provides an easy way to add files and folders to your .gitignore file so that you
don't have to do it by hand! When you run 

    $ gt ignore

Gitutor will display a list with all the files and folders in your current directory. You
only need to check all the files or folders you want to ignore, or uncheck any previously
ignored file so that Git will begin to track it. 

If you check a file or folder which was previously tracked by Git, Gitutor will run:

    $ git rm --cached <filename>

This will remove said file form Git's tracked files and it will also remove the file from
your remote repository, but not from your local one.

'''
    },
}



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['gitutor',
 'gitutor.compare',
 'gitutor.goBack',
 'gitutor.ignore',
 'gitutor.init',
 'gitutor.lesson',
 'gitutor.save']

package_data = \
{'': ['*']}

install_requires = \
['GitPython==3.1.3',
 'click==7.1.2',
 'colorama==0.4.3',
 'gitdb==4.0.5',
 'pygithub==1.51',
 'pyinquirer==1.0.3',
 'requests>=2.20.0,<3.0.0',
 'smmap==3.0.4']

entry_points = \
{'console_scripts': ['gt = gitutor.cli:main']}

setup_kwargs = {
    'name': 'gitutor',
    'version': '0.5.5',
    'description': 'A command line app that makes Git easy.',
    'long_description': "[![pypiv](https://img.shields.io/pypi/v/gitutor.svg)](https://pypi.python.org/pypi/gitutor)\n[![pyv](https://img.shields.io/pypi/pyversions/gitutor.svg)](https://pypi.python.org/pypi/gitutor)\n\n# Gitutor\n\nWelcome to Gitutor. This tool is meant to get you up and running using gitmin the shortest time possible while learning on the go.\n\nGitutor is a command line application that wraps git and provides beginner friendly versions of git's commands. It's Git the easy way.\n\nYou can check out the tutorial and a further explanation of the commands in the [docs](https://gitutor.io/guide). And don't worry if you forget how to use a command you can always run\n\n    $ gt <command> --help\n\n## Available commands\n\n1. gt init - Initialize your local and remote repository.\n2. gt save - Save you changes in the local and remote repository.\n3. gt goback - Return to a previous commit.\n4. gt compare - Compare the current state with a previous commit.\n5. gt ignore - Make git ignore selected files.\n6. gt lesson - See gitutor lessons and documentation.\n\n## Installation guide\n\nIn order to use gitutor without any dependencies version conflicts we recommend installing it using pipx.\n\nPipx creates a virtual environment for your package and exposes its entry point so you can run gitutor from anywhere. To install pipx and configure the $PATH run the following commands:\n\n    $ python3 -m pip install --user pipx\n    $ pipx ensurepath\n\nOnce pipx is installed, run the following to install gitutor\n\n    $ pipx install gitutor\n\nAnd to upgrade gitutor to its latest version you only need to run\n\n    $ pipx upgrade gitutor\n\n## Additional notes\n\nBefore using gitutor you need to have Git available in your computer. You can check the installation guide [here](https://git-scm.com/book/en/v2/Getting-Started-Installing-Git).\n\nIt's also recommended to store your GitHub credentials so you won't have to authenticate everytime you realize a push or pull. You can do this by running\n\n    $ git config --global credential.helper store\n\nThis will store your credentials in a plain-text file (.git-gredentials) under your project directory. If you don't like this you can use any of the following approaches:\n\nOn Mac OS X you can use its native keystore with\n\n    $ git config --global credential.helper oskeychain\n\nFor Windows you can install a helper called [Git Credential Manager for Windows](https://github.com/Microsoft/Git-Credential-Manager-for-Windows) and then run\n\n    $ git config --global credential.helper manager\n",
    'author': 'AMAI',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://gitutor.io',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



import commands
import unittest
from unittest import mock
from click.testing import CliRunner

class TestIgnoreCommand(unittest.TestCase):

    @mock.patch("commands.os")
    def test_repo_without_gitignore(self, os_mock):
        os_mock.path.exists.return_value = False
        gitignore_files = commands.get_gitignore_files("git path")
        self.assertEqual(len(gitignore_files), 0)
        self.assertIsInstance(gitignore_files, list)

    @mock.patch("commands.os")
    @mock.patch("builtins.open", new_callable = mock.mock_open, read_data="\na1\n  \n\na2\na3\n")
    def test_repo_with_gitignore(self, open_mock, os_mock):
        os_mock.path.exists.return_value = True

        gitignore_files = commands.get_gitignore_files("git path")
        self.assertEqual(len(gitignore_files), 3)

    @mock.patch("commands.os")
    def test_get_dir_files(self, os_mock):
        files = ["archivo1", "archivo2", "archivo3"]
        output = {k:{"file_name": f"inside path/{k}", "is_folder": False} for k in files}

        os_mock.listdir.return_value = files
        os_mock.path.isdir.return_value = False

        dir_files = commands.get_files("git path/inside path", "git path")

        self.assertDictEqual(dir_files, output)

    @mock.patch("commands.os")
    def test_get_dir_folders(self, os_mock):
        files = ["folder1", "folder2", "folder3"]
        output = {f"{k}/":{"file_name": f"inside path/{k}/", "is_folder": True} for k in files}

        os_mock.listdir.return_value = files
        os_mock.path.isdir.return_value = True

        dir_files = commands.get_files("git path/inside path", "git path")

        self.assertDictEqual(dir_files, output)

    @mock.patch("commands.os")
    def test_get_dir_files_top(self, os_mock):
        files = ["archivo1", "archivo2", "archivo3"]
        output = {k:{"file_name": k, "is_folder": False} for k in files}

        os_mock.listdir.return_value = files
        os_mock.path.isdir.return_value = False

        dir_files = commands.get_files("git path", "git path")

        self.assertDictEqual(dir_files, output)

    def test_choices(self):
        dir_files = {
            "a1":{
                "file_name": "inside path/a1",
                "is_folder": False
            },
            "a2":{
                "file_name": "inside path/a2",
                "is_folder": False
            },
            "a3":{
                "file_name": "inside path/a3",
                "is_folder": False
            },
            "f1/":{
                "file_name": "inside path/f1/",
                "is_folder": True
            },
            "f2/":{
                "file_name": "inside path/f2/",
                "is_folder": True
            },
            "f3/":{
                "file_name": "inside path/f3/",
                "is_folder": True
            }
        }

        gitignore_files = ["a1", "inside path/a2","inside path/inside/f1/", "inside path/f2/"]

        expected_output = [
            {"name": "a1"},
            {"name": "a2", "checked": True},
            {"name": "a3"},
            {"name": "f1/"},
            {"name": "f2/", "checked": True},
            {"name": "f3/"},
        ]

        choices = commands.create_choices(dir_files, gitignore_files)
        self.assertListEqual(choices, expected_output)

        for key in ["a2", "f2/"]:
            self.assertTrue(dir_files[key]["is_ignored"])

        for key in ["a1", "f1/", "a3", "f3/"]:
            self.assertFalse(dir_files[key]["is_ignored"])
    
    @mock.patch("builtins.open")
    def test_update_gitignore_add_files(self, open_mock):
        dir_files = {
            "a1":{
                "file_name": "inside path/a1",
                "is_folder": False,
                "is_ignored": False
            },
            "a2":{
                "file_name": "inside path/a2",
                "is_folder": False,
                "is_ignored": True
            },
            "a3":{
                "file_name": "inside path/a3",
                "is_folder": False,
                "is_ignored": False
            },
            "f1/":{
                "file_name": "inside path/f1/",
                "is_folder": True,
                "is_ignored": False
            },
            "f2/":{
                "file_name": "inside path/f2/",
                "is_folder": True,
                "is_ignored": True
            },
            "f3/":{
                "file_name": "inside path/f3/",
                "is_folder": True,
                "is_ignored": False
            }
        }
        gitignore_files = ["a1", "inside path/a2","inside path/inside/f1/", "inside path/f2/"]
        answers = ["a1", "a2", "f2/", "f3/"]

        ignored_files = gitignore_files + ["a1", "f3/"]
        ignored_files = "\n".join(ignored_files)

        open_mock = mock.mock_open()

        new_files = commands.update_gitignore(answers, dir_files, gitignore_files, "git path")
        
        self.assertListEqual(new_files, ["a1", "f3/"])

        open_mock.assert_called_once_with('git path/.gitignore', 'w')
        handle = open_mock()
        handle.write.assert_called_once_with(ignored_files)

        

if __name__ == '__main__':
    unittest.main()